// composition-kernel.h
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@cs.nyu.edu (Cyril Allauzen)
//
// \file
// Composition kernel representation.

#ifndef KERNEL_LIB_COMPOSITION_KERNEL_H__
#define KERNEL_LIB_COMPOSITION_KERNEL_H__

#include "kernel/lib/kernel.h"

namespace kernel {


class Polynomial {
 public:
  Polynomial(int degree, double gamma, double coef0) :
      degree_(degree), gamma_(gamma), coef0_(coef0) {}

  double operator()(double x) {
    return powi(gamma_ * x + coef0_, degree_);
  }

  bool IsRbf() { return false; }

  static Polynomial *Read(istream &strm) {
    int64 degree;
    double gamma, coef0;
    ReadType(strm, &degree);
    ReadType(strm, &gamma);
    ReadType(strm, &coef0);
    return new Polynomial(degree, gamma, coef0);
  }

  void Write(ostream &strm) {
    WriteType(strm, (int64) degree_);
    WriteType(strm, gamma_);
    WriteType(strm, coef0_);
  }

  static const string &Type() {
    static const string type = "polynomial";
    return type;
  }

  uint64 Properties(uint64 inprops) {
    uint64 outprops = 0;
    if (inprops & (kSymmetric | kAsymmetric))
      outprops |= inprops & (kSymmetric | kAsymmetric);
    if ((gamma_ > 0) && (coef0_ >= 0) && inprops & kPosDefinite)
      outprops |= kPosDefinite;
    return outprops;
  }

 private:
  inline double powi(double base, int times) {
    double tmp = base, ret = 1.0;
    for (int t = times; t > 0; t /= 2) {
      if (t % 2 == 1) ret *= tmp;
      tmp = tmp * tmp;
    }
    return ret;
  }

  int degree_;
  double gamma_;
  double coef0_;
};


class Sigmoid {
 public:
  Sigmoid(double gamma, double coef0) :
      gamma_(gamma), coef0_(coef0) {}

  double operator()(double x) {
    return tanh(gamma_ * x + coef0_);
  }

  bool IsRbf() { return false; }

  static Sigmoid *Read(istream &strm) {
    double gamma, coef0;
    ReadType(strm, &gamma);
    ReadType(strm, &coef0);
    return new Sigmoid(gamma, coef0);
  }

  void Write(ostream &strm) {
    WriteType(strm, gamma_);
    WriteType(strm, coef0_);
  }

  static const string &Type() {
    static const string type = "sigmoid";
    return type;
  }

  uint64 Properties(uint64 inprops) {
    uint64 outprops = 0;
    if (inprops & (kSymmetric | kAsymmetric))
      outprops |= inprops & (kSymmetric | kAsymmetric);
    if ((gamma_ > 0) && (coef0_ >= 0) && inprops & kPosDefinite)
      outprops |= kPosDefinite;
    return outprops;
  }

 private:
  double gamma_;
  double coef0_;
};


class Gaussian {
 public:
  Gaussian(double gamma) :
      gamma_(gamma) {}

  double operator()(double x) {
    return exp(-gamma_ * x);
  }

  bool IsRbf() { return true; }

  static Gaussian *Read(istream &strm) {
    double gamma;
    ReadType(strm, &gamma);
    return new Gaussian(gamma);
  }

  void Write(ostream &strm) {
    WriteType(strm, gamma_);
  }

  static const string &Type() {
    static const string type = "gaussian";
    return type;
  }

  uint64 Properties(uint64 inprops) {
    uint64 outprops = 0;
    if (inprops & (kSymmetric | kAsymmetric))
      outprops |= inprops & (kSymmetric | kAsymmetric);
    if ((gamma_ > 0) && inprops & kPosDefinite)
      outprops |= kPosDefinite;
    return outprops;
  }

 private:
  double gamma_;
};



template <class O, class F>
class CompositionKernelImpl : public KernelImpl<O> {
 public:
  CompositionKernelImpl(Kernel<O> *kernel, F *function) :
      kernel_(kernel->Copy()), function_(function) {
    KernelImpl<O>::SetProperties(function_->Properties(kernel->Properties()));
  }

  ~CompositionKernelImpl() {
    VLOG(1) << "~CompositionKernelImpl[]";
    delete kernel_;
    delete function_;
  }

  double Eval(size_t i, size_t j) {
    if (function_->IsRbf())
      return (*function_)(kernel_->Eval(i,i) + kernel_->Eval(j,j)
                          - 2 * kernel_->Eval(i,j));
    else
      return (*function_)(kernel_->Eval(i,j));
  }

  void SetData(DataSet<O> *ds) { kernel_->SetData(ds);}

  const DataSet<O> *Data() const { return kernel_->Data();}

  DataSet<O> *Data() { return kernel_->Data();}

  static CompositionKernelImpl<O, F> *Read(istream &strm,
                                           const KernelReadOptions &opts) {
    CompositionKernelImpl<O, F> *impl = new CompositionKernelImpl<O, F>();
    impl->SetType(F::Type());
    KernelHeader hdr;
    if (!impl->ReadHeader(strm, opts, 0, &hdr))
      return 0;
    impl->function_ = F::Read(strm);
    if (!impl->function_) {
      LOG(ERROR) << "CompositionKernel::Read: read failed: " << opts.source;
      return 0;
    }
    impl->kernel_ = Kernel<O>::Read(strm, KernelReadOptions(opts.source));
    if (!impl->kernel_) {
      LOG(ERROR) << "CompositionKernel::Read: read failed: " << opts.source;
      return 0;
    }
    return impl;
  }

  bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    KernelHeader hdr;
    KernelImpl<O>::WriteHeader(strm, opts, 0, &hdr);
    if (!strm)
      return false;
    function_->Write(strm);
    VLOG(1) << "CompositionKernelImpl::Write: writing argument kernel "
            << kernel_->Type();
    kernel_->Write(strm, KernelWriteOptions(opts.source));
    if (!strm)
      LOG(ERROR) << "CompositionKernel::Write: write failed: " << opts.source;
    return strm;
  }

 private:
  CompositionKernelImpl() : kernel_(0), function_(0) {}

 private:
  Kernel<O> *kernel_;
  F *function_;
};


template <class O>
class PolynomialKernel : public Kernel<O> {
 public:
  typedef O Object;
  typedef CompositionKernelImpl<O, Polynomial> Impl;

  PolynomialKernel(Kernel<O> *kernel, int degree, double gamma, double coef0)
      : impl_(new Impl(kernel, new Polynomial(degree, gamma, coef0))) {
    impl_->SetType("polynomial");
  }

  PolynomialKernel(const PolynomialKernel<O> &kernel) : impl_(kernel.impl_) {
    impl_->IncrRefCount();
  }

  virtual ~PolynomialKernel() {
    if(!impl_->DecrRefCount()) delete impl_;
  }

  PolynomialKernel<O> &operator=(const PolynomialKernel<O> &kernel) {
    if (this != &kernel) {
      if(!impl_->DecrRefCount()) delete impl_;
      kernel.impl_->IncrRefCount();
      impl_ = kernel.impl_;
    }
    return *this;
  }

  virtual double Eval(size_t i, size_t j) {
    return impl_->Eval(i,j);
  }

  virtual const string& Type() const { return impl_->Type(); }

  virtual PolynomialKernel<O> *Copy() const {
    impl_->IncrRefCount();
    return new PolynomialKernel<O>(impl_);
  }

  static PolynomialKernel<O> *Read(istream &strm,
                                   const KernelReadOptions &opts) {
    Impl *impl = Impl::Read(strm, opts);
    impl->SetType("polynomial");
    return new PolynomialKernel(impl);
  }

  static PolynomialKernel<O> *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "PolynomialKernel::Read: Can't open file: " << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

  virtual bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    impl_->Write(strm, opts);
    return strm;
  }

  virtual bool Write(const string &filename) const {
    if (!filename.empty()) {
      ofstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "PolynomialKernel::Write: Can't open file: " << filename;
        return false;
      }
      return Write(strm, KernelWriteOptions(filename));
    } else {
      return Write(std::cout, KernelWriteOptions("standard output"));
    }
  }

  virtual void SetData(DataSet<O>* ds) {
    impl_->SetData(ds);
  }

  virtual const DataSet<O>* Data() const {
    return impl_->Data();
  }

  virtual DataSet<O>* Data() {
    return impl_->Data();
  }

  virtual uint64 Properties(uint64 mask) const {
    return impl_->Properties(mask);
  }

  virtual void SetProperties(uint64 props, uint64 mask) {
    impl_->SetProperties(props, mask);
  }

 protected:
  explicit PolynomialKernel(Impl *impl) : impl_(impl) {}

 private:
  Impl *impl_;
};

template <class O>
class SigmoidKernel : public Kernel<O> {
 public:
  typedef O Object;
  typedef CompositionKernelImpl<O, Sigmoid> Impl;

  SigmoidKernel(Kernel<O> *kernel, double gamma, double coef0)
      : impl_(new Impl(kernel, new Sigmoid(gamma, coef0))) {
    impl_->SetType("sigmoid");
  }

  SigmoidKernel(const SigmoidKernel<O> &kernel) : impl_(kernel.impl_) {
    impl_->IncrRefCount();
  }

  virtual ~SigmoidKernel() {
    if(!impl_->DecrRefCount()) delete impl_;
  }

  SigmoidKernel<O> &operator=(const SigmoidKernel<O> &kernel) {
    if (this != &kernel) {
      if(!impl_->DecrRefCount()) delete impl_;
      kernel.impl_->IncrRefCount();
      impl_ = kernel.impl_;
    }
    return *this;
  }

  virtual double Eval(size_t i, size_t j) {
    return impl_->Eval(i,j);
  }

  virtual const string& Type() const { return impl_->Type(); }

  virtual SigmoidKernel<O> *Copy() const {
    impl_->IncrRefCount();
    return new SigmoidKernel<O>(impl_);
  }

  static SigmoidKernel<O> *Read(istream &strm,
                                   const KernelReadOptions &opts) {
    Impl *impl = Impl::Read(strm, opts);
    impl->SetType("sigmoid");
    return new SigmoidKernel(impl);
  }

  static SigmoidKernel<O> *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "SigmoidKernel::Read: Can't open file: " << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

  virtual bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    impl_->Write(strm, opts);
    return strm;
  }

  virtual bool Write(const string &filename) const {
    if (!filename.empty()) {
      ofstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "SigmoidKernel::Write: Can't open file: " << filename;
        return false;
      }
      return Write(strm, KernelWriteOptions(filename));
    } else {
      return Write(std::cout, KernelWriteOptions("standard output"));
    }
  }

  virtual void SetData(DataSet<O>* ds) {
    impl_->SetData(ds);
  }

  virtual const DataSet<O>* Data() const {
    return impl_->Data();
  }

  virtual DataSet<O>* Data() {
    return impl_->Data();
  }

  virtual uint64 Properties(uint64 mask) const {
    return impl_->Properties(mask);
  }

  virtual void SetProperties(uint64 props, uint64 mask) {
    impl_->SetProperties(props, mask);
  }

 protected:
  explicit SigmoidKernel(Impl *impl) : impl_(impl) {}

 private:
  Impl *impl_;
};

template <class O>
class GaussianKernel : public Kernel<O> {
 public:
  typedef O Object;
  typedef CompositionKernelImpl<O, Gaussian> Impl;

  GaussianKernel(Kernel<O> *kernel, double gamma)
      : impl_(new Impl(kernel, new Gaussian(gamma))) {
    impl_->SetType("gaussian");
  }

  GaussianKernel(const GaussianKernel<O> &kernel) : impl_(kernel.impl_) {
    impl_->IncrRefCount();
  }

  virtual ~GaussianKernel() {
    if(!impl_->DecrRefCount()) delete impl_;
  }

  GaussianKernel<O> &operator=(const GaussianKernel<O> &kernel) {
    if (this != &kernel) {
      if(!impl_->DecrRefCount()) delete impl_;
      kernel.impl_->IncrRefCount();
      impl_ = kernel.impl_;
    }
    return *this;
  }

  virtual double Eval(size_t i, size_t j) {
    return impl_->Eval(i,j);
  }

  virtual const string& Type() const { return impl_->Type(); }

  virtual GaussianKernel<O> *Copy() const {
    impl_->IncrRefCount();
    return new GaussianKernel<O>(impl_);
  }

  static GaussianKernel<O> *Read(istream &strm,
                                   const KernelReadOptions &opts) {
    Impl *impl = Impl::Read(strm, opts);
    impl->SetType("gaussian");
    return new GaussianKernel(impl);
  }

  static GaussianKernel<O> *Read(const string &filename) {
    if (!filename.empty()) {
      ifstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "GaussianKernel::Read: Can't open file: " << filename;
        return 0;
      }
      return Read(strm, KernelReadOptions(filename));
    } else {
      return Read(std::cin, KernelReadOptions("standard input"));
    }
  }

  virtual bool Write(ostream &strm, const KernelWriteOptions &opts) const {
    impl_->Write(strm, opts);
    return strm;
  }

  virtual bool Write(const string &filename) const {
    if (!filename.empty()) {
      ofstream strm(filename.c_str());
      if (!strm) {
        LOG(ERROR) << "GaussianKernel::Write: Can't open file: " << filename;
        return false;
      }
      return Write(strm, KernelWriteOptions(filename));
    } else {
      return Write(std::cout, KernelWriteOptions("standard output"));
    }
  }

  virtual void SetData(DataSet<O>* ds) {
    impl_->SetData(ds);
  }

  virtual const DataSet<O>* Data() const {
    return impl_->Data();
  }

  virtual DataSet<O>* Data() {
    return impl_->Data();
  }

  virtual uint64 Properties(uint64 mask) const {
    return impl_->Properties(mask);
  }

  virtual void SetProperties(uint64 props, uint64 mask) {
    impl_->SetProperties(props, mask);
  }

 protected:
  explicit GaussianKernel(Impl *impl) : impl_(impl) {}

 private:
  Impl *impl_;
};


}  // namespace kernel

#endif  // KERNEL_LIB_COMPOSITION_KERNEL
