
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#include <stdarg.h>

#include "kernel/lib/liblinear.h"

// This file contains functions that are copied directly from LIBLINEAR,
// and modified only to also return the dual optimal variables (i.e. the
// double *alpha parameter in solve_linear_c_svc).

namespace kernel {

#define INF HUGE_VAL

static void info(const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  vprintf(fmt,ap);
  va_end(ap);
}

static void info_flush()
{
  fflush(stdout);
}

void solve_linear_c_svc(const problem *prob, double *w, double *alpha,
                               double eps, double Cp, double Cn,
                               double *f, // linear term coefficients
                               int solver_type) {


  int l = prob->l;
  int n = prob->n;
  int i, s, iter = 0;
  double C, d, G;
  double *QD = new double[l];
  int max_iter = 20000;
  int *index = new int[l];
  //double *alpha = new double[l]; //rostami: allocated outside of function
  schar *y = new schar[l];
  int active_size = l;

  // PG: projected gradient, for shrinking and stopping
  double PG;
  double PGmax_old = INF;
  double PGmin_old = -INF;
  double PGmax_new, PGmin_new;

  // default solver_type: L2LOSS_SVM_DUAL
  double diag_p = 0.5/Cp, diag_n = 0.5/Cn;
  double upper_bound_p = INF, upper_bound_n = INF;
  if(solver_type == L1LOSS_SVM_DUAL)
  {
    diag_p = 0; diag_n = 0;
    upper_bound_p = Cp; upper_bound_n = Cn;
  }

  for(i=0; i<n; i++)
    w[i] = 0;
  for(i=0; i<l; i++)
  {
    alpha[i] = 0;
    if(prob->y[i] > 0)
    {
      y[i] = +1;
      QD[i] = diag_p;
    }
    else
    {
      y[i] = -1;
      QD[i] = diag_n;
    }

    feature_node *xi = prob->x[i];
    while (xi->index != -1)
    {
      QD[i] += (xi->value)*(xi->value);
      xi++;
    }
    index[i] = i;
  }

  while (iter < max_iter)
  {
    PGmax_new = -INF;
    PGmin_new = INF;

    for (i=0; i<active_size; i++)
    {
      int j = i+rand()%(active_size-i);
      swap(index[i], index[j]);
    }

    for (s=0;s<active_size;s++)
    {
      i = index[s];
      G = 0;
      schar yi = y[i];

      feature_node *xi = prob->x[i];
      while(xi->index!= -1)
      {
        G += w[xi->index-1]*(xi->value);
        xi++;
      }
      //G = G*yi-1;
      G = G*yi+f[i];

      if(yi == 1)
      {
        C = upper_bound_p;
        G += alpha[i]*diag_p;
      }
      else
      {
        C = upper_bound_n;
        G += alpha[i]*diag_n;
      }

      PG = 0;
      if (alpha[i] == 0)
      {
        if (G > PGmax_old)
        {
          active_size--;
          swap(index[s], index[active_size]);
          s--;
          continue;
        }
        else if (G < 0)
          PG = G;
      }
      else if (alpha[i] == C)
      {
        if (G < PGmin_old)
        {
          active_size--;
          swap(index[s], index[active_size]);
          s--;
          continue;
        }
        else if (G > 0)
          PG = G;
      }
      else
        PG = G;

      PGmax_new = max(PGmax_new, PG);
      PGmin_new = min(PGmin_new, PG);

      if(fabs(PG) > 1.0e-12)
      {
        double alpha_old = alpha[i];
        alpha[i] = min(max(alpha[i] - G/QD[i], 0.0), C);
        d = (alpha[i] - alpha_old)*yi;
        xi = prob->x[i];
        while (xi->index != -1)
        {
          w[xi->index-1] += d*xi->value;
          xi++;
        }
      }
    }

    iter++;
    if(iter % 1000 == 0)
    {
//      info(".");
      info_flush();
      printf("PGmax: %g,  PGmin: %g\n", PGmax_new, PGmin_new);
      printf("active_size: %d\n", active_size);
    }

    if(PGmax_new - PGmin_new <= eps)
    {
      if(active_size == l)
        break;
      else
      {
        active_size = l;
        //info("*"); info_flush();
        PGmax_old = INF;
        PGmin_old = -INF;
        continue;
      }
    }
    PGmax_old = PGmax_new;
    PGmin_old = PGmin_new;
    if (PGmax_old <= 0)
      PGmax_old = INF;
    if (PGmin_old >= 0)
      PGmin_old = -INF;
  }

  info("\noptimization finished, #iter = %d\n",iter);
  if (iter >= max_iter)
    info("Warning: reaching max number of iterations\n");

  // calculate objective value

  double v = 0;
  int nSV = 0;
  for(i=0; i<n; i++)
    v += w[i]*w[i];
  for(i=0; i<l; i++)
  {
    if (y[i] == 1)
      v += alpha[i]*(alpha[i]*diag_p - 2);
    else
      v += alpha[i]*(alpha[i]*diag_n - 2);
    if(alpha[i] > 0)
      ++nSV;
  }
  info("Objective value = %lf\n",v/2);
  info("nSV = %d\n",nSV);

  delete [] QD;
  //delete [] alpha; //rostami deleted outside of function.
  delete [] y;
  delete [] index;
}

} // namespace kernel
