
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB_LIBSVM_H_
#define KERNEL_LIB_LIBSVM_H_

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/semidense-matrix.h"
#include "kernel/lib/dense-vector.h"
#include "kernel/lib/sparse-vector.h"

#include "libsvm-2.82/svm.h"


// Explicit definition of svm_model copied from svm.cpp,
// needed by LibSVMTrain below.
struct svm_model
{
  svm_parameter param;  // parameter
  int nr_class;      // number of classes, = 2 in regression/one class svm
  int l;             // total #SV
  svm_node **SV;     // SVs (SV[l])
  double **sv_coef;  // coefficients for SVs in decision
                     // functions (sv_coef[n-1][l])
  double *rho;       // constants in decision functions (rho[n*(n-1)/2])
  double *probA;     // pairwise probability information
  double *probB;

  // for classification only

  int *label;    // label of each class (label[n])
  int *nSV;    // number of SVs for each class (nSV[n])
               // nSV[0] + nSV[1] + ... + nSV[n-1] = l
  int free_sv;    // 1 if svm_model is created by svm_load_model
                  // 0 if svm_model is created by svm_train
};

namespace kernel {

// OpenKernel wrapper function.
// Solves the l1-loss svm problem, given the kernel matrix.
// // INPUT:
// kernel - a Dense matrix containing the kernel matrix.
// labels - a DenseVector of labels for each instance,
// reg_param - the SVM regularization parameter "C", which upper bounds
//   the size of each dual variable.
// OUTPUT:
// dual_vars - hold the optimal dual variables
template <typename T>
void LibSVMTrain(const DenseMatrix<T> &kernel,
                    const DenseVector<T> &labels,
                    const double &reg_param,
                    DenseVector<T> *dual_vars,
                    double *offset) {
  svm_problem prob;
  prob.l = kernel.NumRow();
  prob.y = new double[prob.l];
  for (size_t i = 0; i < prob.l; ++i)
    prob.y[i] = labels.Element(i);
  prob.x = new struct svm_node*[prob.l];
  KernelToLibSVM(kernel, prob.x);

  svm_parameter param;
  param.svm_type = C_SVC;
  param.kernel_type = PRECOMPUTED;
  param.cache_size = 100;
  param.eps = 0.001;
  param.C = reg_param;
  param.nr_weight = 0;
  param.shrinking = 0;
  param.probability = 0;

  svm_model *model = svm_train(&prob, &param);

  dual_vars->Clear();
  dual_vars->Resize(prob.l, 0);
  VLOG(2) << "nr_class: " << model->nr_class
            << " num_pos: " <<  model->nSV[0]
            << " num_neg: " <<  model->nSV[1];
  //positive class
  for (size_t i = 0; i < model->nSV[0]; ++i)
    dual_vars->SetElement(static_cast<int>(model->SV[i][0].value) - 1,
                          abs(model->sv_coef[0][i]));
  //negative class
  for (size_t i = 0; i < model->nSV[1]; ++i)
    dual_vars->SetElement(
                  static_cast<int>(model->SV[i + model->nSV[0]][0].value) - 1,
                  abs(model->sv_coef[0][i + model->nSV[0]]));

  *offset = -model->rho[0];
  if (model->label[0] == -1)
    *offset = -*offset;

  for (size_t i = 0; i < prob.l; ++i)
    delete[] prob.x[i];
  delete[] prob.x;
  delete[] prob.y;
  svm_destroy_model(model);
}

// This specializtion should never be used, only defined so that 
// kernel-learning-linearl1 compliles.
// TODO(rostami): a more elegant solution?
template <typename T, typename M>
void LibSVMTrain(const M &kernel,
                    const DenseVector<T> &labels,
                    const double &reg_param,
                    DenseVector<T> *dual_vars,
                    double *offset) {
  LOG(FATAL) << "LibSVMTrain non-DenseMatrix specialization used";
}


// Convert an OpenKernel DenseMatrix to a precomputed libsvm kernel
template <typename T> inline
void KernelToLibSVM(const DenseMatrix<T> &kernel, svm_node **x) {
  for (size_t i = 0; i < kernel.NumRow(); ++i) {
    x[i] = new struct svm_node[kernel.NumColumn() + 2];
    // index 0 stores point id
    x[i][0].index = 0;
    x[i][0].value = i + 1;
    for (size_t j = 0; j < kernel.NumColumn(); ++j) {
      x[i][j+1].index = j + 1;
      x[i][j+1].value = kernel.Element(i,j);
    }
    x[i][kernel.NumColumn()+1].index = -1;
  }
}

template <typename T> inline
void SVMDualPredict(const DenseMatrix<T> &kernel, const DenseVector<T> &alpha,
                const DenseVector<T> &labels, double offset,
                DenseVector<T> *pred) {
  pred->Clear();
  pred->Resize(kernel.NumRow(), offset);
  DenseVector<T> yalpha;
  PointwiseProduct(alpha, labels, &yalpha);
  MatrixVectorProduct<T>(1.0, kernel, yalpha, 1.0, pred);
  SignVector(*pred, pred);
}

} // namespace kernel

#endif  // KERNEL_LIB_LIBSVM_H_
