// semidense-matrix.h
//

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@google.com (Cyril Allauzen)
//
// \file
// A simple reference-counted representation for matrices
// dense along at least one dimension


#ifndef KERNEL_LIB_SEMI_DENSE_MATRIX_H_
#define KERNEL_LIB_SEMI_DENSE_MATRIX_H_

#include <vector>

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/sparse-vector.h"
#include "kernel/lib/kernel.h" // change to min req

namespace kernel {

//
// SEMI-DENSE MATRIX CLASS
//

enum SemiDenseType { DENSE_BY_ROW = 0, DENSE_BY_COLUMN = 1};

#define TRANSPOSE_SEMI_DENSE_TYPE(S)  \
  ((S) == DENSE_BY_ROW ? DENSE_BY_COLUMN : DENSE_BY_ROW)

// Forward declaration
template <class V>
class SemiDenseMatrixData;
template <class V, SemiDenseType S>
class SemiDenseMatrix;

// SemiDenseMatrix: a simple reference-counted representation for
// matrices dense along at least one dimension, templated on the
// vector type used to represents rows ('transpose_ = false') or
// columns ('transpose_ = true')
template <class V, SemiDenseType S>
class SemiDenseMatrixBase {
 public:
  typedef typename V::ElementType ElementType;
  typedef SemiDenseMatrixBase<V, TRANSPOSE_SEMI_DENSE_TYPE(S)> TransposeMatrix;
  friend class SemiDenseMatrixBase<V, TRANSPOSE_SEMI_DENSE_TYPE(S)>;

  SemiDenseMatrixBase() : data_(new SemiDenseMatrixData<V>) {}

  SemiDenseMatrixBase(const SemiDenseMatrixBase<V, S> &base)
      : data_(base.data_) {
       data_->IncrRefCount();
      }

  virtual ~SemiDenseMatrixBase() {
    if (!data_->DecrRefCount())
      delete data_;
  }

  SemiDenseMatrixBase<V, S> &operator=(const SemiDenseMatrixBase<V, S> &m) {
    if (this != &m) {
      if (!data_->DecrRefCount()) delete data_;
      m.data_->IncrRefCount();
      data_ = m.data_;
    }
    return *this;
  }

  void Transpose(TransposeMatrix *m) const {
    *m = TransposeMatrix(data_);
  }

  const V &Vector(size_t i) const { return data_->Vector(i); }
  V *MutableVector(size_t i) { return data_->MutableVector(i); }

  size_t Size() const { return data_->Size(); }

  size_t MaxSparseIndex() const { return data_->MaxSparseIndex(); }

  void AddVector(const V &vec) { data_->AddVector(vec); }

  void Clear() { data_->Clear(); }

 protected:
  SemiDenseMatrixBase(SemiDenseMatrixData<V> *data) : data_(data) {
    data_->IncrRefCount();
  }

  void MutateCheck() {
    // Copy on write
    if (data_->RefCount() > 1) {
      data_->DecrRefCount();
      data_ = new SemiDenseMatrixData<V>(*data_);
    }
  }

 private:
  SemiDenseMatrixData<V> *data_;
};


template <class V>
class SemiDenseMatrix<V, DENSE_BY_ROW>
    : public SemiDenseMatrixBase<V, DENSE_BY_ROW> {
 public:
  typedef SemiDenseMatrix<V, DENSE_BY_COLUMN> TransposeMatrix;
  using SemiDenseMatrixBase<V, DENSE_BY_ROW>::Transpose;
  using SemiDenseMatrixBase<V, DENSE_BY_ROW>::operator=;
  using SemiDenseMatrixBase<V, DENSE_BY_ROW>::Size;
  using SemiDenseMatrixBase<V, DENSE_BY_ROW>::MaxSparseIndex;
  using SemiDenseMatrixBase<V, DENSE_BY_ROW>::Vector;

  SemiDenseMatrix() {};
  SemiDenseMatrix(const SemiDenseMatrixBase<V, DENSE_BY_ROW> &base)
      : SemiDenseMatrixBase<V, DENSE_BY_ROW>(base) {}

  size_t NumRow() const { return Size(); }
  size_t NumColumn() const { return MaxSparseIndex() + 1; }
  const V &Row(size_t i) const { return Vector(i); }
  void AddRow(const V &row) { AddVector(row); }

  // Expensive operations, should not be used in efficient applications
  typename V::ElementType Element(size_t i, size_t j) {
    return Vector(i).Element(j);
  }
  void SetElement(size_t i, size_t j, typename V::ElementType val) {
    this->MutableVector(i)->SetElement(j, val);
  }

};


template <class V>
class SemiDenseMatrix<V, DENSE_BY_COLUMN>
    : public SemiDenseMatrixBase<V, DENSE_BY_COLUMN> {
 public:
  typedef SemiDenseMatrix<V, DENSE_BY_ROW> TransposeMatrix;
  using SemiDenseMatrixBase<V, DENSE_BY_COLUMN>::Transpose;
  using SemiDenseMatrixBase<V, DENSE_BY_COLUMN>::operator=;
  using SemiDenseMatrixBase<V, DENSE_BY_COLUMN>::Size;
  using SemiDenseMatrixBase<V, DENSE_BY_COLUMN>::MaxSparseIndex;
  using SemiDenseMatrixBase<V, DENSE_BY_COLUMN>::Vector;

  SemiDenseMatrix() {};
  SemiDenseMatrix(const SemiDenseMatrixBase<V, DENSE_BY_COLUMN> &base)
      : SemiDenseMatrixBase<V, DENSE_BY_COLUMN>(base) {}

  size_t NumColumn() const { return Size(); }
  size_t NumRow() const { return MaxSparseIndex() + 1; }
  const V &Column(size_t i) const { return Vector(i); }
  void AddColumn(const V &col) { AddVector(col); }
};


// Provides the data representation of a dense matrix as
// an STL vector of vectors of type V.
template <class V>
class SemiDenseMatrixData {
 public:
  SemiDenseMatrixData()
      : max_sparse_index_(0) {}

  explicit SemiDenseMatrixData(const SemiDenseMatrixData<V> &data)
      : vectors_(data.vectors_), max_sparse_index_(0) {}

  ~SemiDenseMatrixData() {}

  size_t Size() const { return vectors_.size(); }
  size_t MaxSparseIndex() const { return max_sparse_index_; }
  void Clear() { vectors_.clear(); }
  const V &Vector(size_t i) const { return vectors_[i]; }
  V *MutableVector(size_t i) { return &vectors_[i]; }
  void AddVector(const V &v) {
    vectors_.push_back(v);
    if (v.MaxIndex() > max_sparse_index_)
      max_sparse_index_ = v.MaxIndex();
  }

  int RefCount() const { return ref_count_.count(); }
  int IncrRefCount() { return ref_count_.Incr(); }
  int DecrRefCount() { return ref_count_.Decr(); }

 private:
  vector<V> vectors_;
  RefCounter ref_count_;
  size_t max_sparse_index_;

  void operator=(const SemiDenseMatrixData<V> &);  // disallow
};


//
// MATRIX OPERATIONS
//

// Copy a subset of vectors along the dense dimension
template <class V, SemiDenseType S, typename I> inline
void SubMatrixCopy(const SemiDenseMatrix<V, S> &mat,
                   const DenseVector<I> &vector_index,
                   SemiDenseMatrix<V, S> *submatrix) {
  submatrix->Clear();
  for (size_t i = 0; i < vector_index.Size(); ++i) {
    CHECK(vector_index.Element(i) < mat.Size());
    submatrix->AddVector(mat.Vector(vector_index.Element(i)));
  }
}

// Return a vector of sparse vectors representing the columns of mat
// (which is dense by row).
template <typename T>
void MatrixToColumnVectors(
                  const SemiDenseMatrix< SparseVector<T>, DENSE_BY_ROW> &mat,
                  vector< SparseVector<T> > *columns) {
  columns->resize(mat.NumColumn());
  SparseVector<T> row;
  for (size_t i = 0; i < mat.NumRow(); ++i) {
    row = mat.Row(i);
    for (size_t j = 0; j < row.Size(); ++j)
      (*columns)[row.Pair(j).first].AddElement(i, row.Pair(j).second);
  }
}

// Simple print function
template <typename T, SemiDenseType S> inline
void PrintSemiDense(const SemiDenseMatrixBase<SparseVector<T>, S> &mat,
                    ostream &out) {
  SparseVector<T> vec;
  for (size_t i = 0; i < mat.Size(); ++i) {
    vec = mat.Vector(i);
    for (size_t j = 0; j < vec.Size(); ++j) {
      pair<size_t, T> p = vec.Pair(j);
      out << "(" << i << "," << p.first << "): " << p.second << endl;
    }
  }
}

// Construct a sparse diagonal matrix, with diagonal equal to
// diag_vector;
template <class V, SemiDenseType S, typename T> inline
void Diagonal(const DenseVector<T> &diag_vector,
              SemiDenseMatrixBase<V, S> *I) {
  for (size_t i = 0; i < diag_vector.Size(); ++i) {
    V temp;
    temp.AddElement(i, diag_vector.Element(i));
    I->AddVector(temp);
  }
}

// TODO(allauzen): changed in a lot of the definition the argument from
// SemiDenseMatrixBase to SemiDenseMatrix in order to avoid conflict
// with the templates in sparse-vector.h. Need to investigate
// a more elegant fix for this.

// Equality tests
template <class V, SemiDenseType S> inline
bool Equal(const SemiDenseMatrix<V, S> &a,
           const SemiDenseMatrix<V, S> &b) {
  if (a.Size() != b.Size())
    return false;

  bool flag = true;
  for (size_t i = 0; i < a.Size(); ++i)
    if (!Equal(a.Vector(i), b.Vector(i)))
      flag = false;

  return flag;
}

template <class V, SemiDenseType S> inline
bool ApproxEqual(const SemiDenseMatrix<V, S> &a,
                 const SemiDenseMatrix<V, S> &b,
                 typename V::ElementType delta) {
  if (a.Size() != b.Size())
    return false;

  bool flag = true;
  for (size_t i = 0; i < a.Size(); ++i)
    if (!ApproxEqual(a.Vector(i), b.Vector(i), delta))
      flag = false;

  return flag;
}

template <class V, SemiDenseType S> inline
bool ApproxEqual(const SemiDenseMatrix<V, S> &a,
                 const DenseMatrix<typename V::ElementType> &b_dense,
                 typename V::ElementType delta) {

  SemiDenseMatrix<V, S> b;
  DenseToSemiDense(b_dense, &b);

  return ApproxEqual(a, b, delta);
}

// Convert a DenseMatrix to a SemiDenseMatrix.
// Used to compare dense and semidense matrices.
template <class V> inline
void DenseToSemiDense(const DenseMatrix<typename V::ElementType> &dense,
                      SemiDenseMatrixBase<V, DENSE_BY_ROW> *sdense) {
  sdense->Clear();
  for (size_t i = 0; i < dense.NumRow(); ++i) {
    V row;
    for (size_t j = 0; j < dense.NumColumn(); ++j) {
      row.AddElement(j, dense.Element(i, j));
    }
    sdense->AddVector(row);
  }
}

template <class V> inline
void DenseToSemiDense(const DenseMatrix<typename V::ElementType> &dense,
                      SemiDenseMatrixBase<V, DENSE_BY_COLUMN> *sdense) {
  sdense->Clear();
  for (size_t j = 0; j < dense.NumColumn(); ++j) {
    V col;
    for (size_t i = 0; i < dense.NumRow(); ++i) {
      col.AddElement(i, dense.Element(i, j));
    }
    sdense->AddVector(col);
  }
}

// C = A * B
// with A dense by row and B dense by column
template <typename V> inline
void Product(const SemiDenseMatrix<V, DENSE_BY_ROW> &a,
             const SemiDenseMatrix<V, DENSE_BY_COLUMN> &b,
             DenseMatrix<typename V::ElementType> *c) {
  c->Reset(a.NumRow(), b.NumColumn(), 0.0F);
  for (size_t i = 0; i < a.NumRow(); ++i)
    for (size_t j = 0; j < b.NumColumn(); ++j )
      c->SetElement(i, j, DotProduct(a.Row(i), b.Column(j)));
}

// C = A * B
// with A dense by row and B dense by column
// return a semidense matrix
template <typename V> inline
void Product(const SemiDenseMatrix<V, DENSE_BY_ROW> &a,
             const SemiDenseMatrix<V, DENSE_BY_COLUMN> &b,
             SemiDenseMatrix<V, DENSE_BY_ROW> *c) {
  c->Clear();
  for (size_t i = 0; i < a.NumRow(); ++i) {
    V temp;
    for (size_t j = 0; j < b.NumColumn(); ++j ) {
      temp.AddElement(j, DotProduct(a.Row(i), b.Column(j)));
    }
    c->AddRow(temp);
  }
}

// C = A * B
// with A dense by column and B dense by row
template <typename T> inline
void Product(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_COLUMN> &a,
             const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> &b,
             DenseMatrix<T> *c) {

  CHECK(a.NumColumn() == b.NumRow());
  c->Reset(a.NumRow(), b.NumColumn(), 0);
  for (size_t k = 0; k < a.NumColumn() ; ++k) {
    const SparseVector<T> &column = a.Column(k);
    const SparseVector<T> &row = b.Row(k);
    for (size_t i = 0; i < column.Size(); ++i)
      for (size_t j = 0; j < row.Size(); ++j) {
        pair<size_t, T>
            pair_i = column.Pair(i), pair_j = row.Pair(j);
        c->SetElement(pair_i.first, pair_j.first,
                      c->Element(pair_i.first, pair_j.first) +
                      pair_i.second * pair_j.second);
      }
  }
}

//
// MATRIX-VECTOR OPERATIONS
//

// B = A * diag(v)
template <typename T> inline
void DiagonalProduct(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> &A,
                     const DenseVector<T> &v,
                     SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> *B) {
  B->Clear();
  for (size_t i = 0; i < A.NumRow(); ++i) {
    SparseVector<T> temp;
    PointwiseProduct(A.Row(i), v, &temp);
    B->AddRow(temp);
  }
}

// u = B * v
// B is dense by column.
template <typename T> inline
void Product(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_COLUMN> &a,
             const DenseVector<T> &v,
             DenseVector<T> *u) {
  CHECK(v.Size() == a.NumColumn());
  u->Clear();
  if (u->Size() < a.NumRow())
    u->Resize(a.NumRow(), 0);

  SparseVector<T> col;
  pair<size_t, T> p;
  for (size_t j = 0; j < a.NumColumn(); ++j) {
    col = a.Column(j);
    for (size_t i = 0; i < col.Size(); ++i) {
      p = col.Pair(i);
      u->SetElement(p.first, u->Element(p.first)
                    + v.Element(j) * p.second);
    }
  }
}

// u = B * v
// B is dense by row.
template <typename T> inline
void Product(const SemiDenseMatrix<SparseVector<T>, DENSE_BY_ROW> &a,
             const DenseVector<T> &v,
             DenseVector<T> *u) {
  if (u->Size() != a.NumRow())
    u->Resize(a.NumRow(), 0);

  SparseVector<T> row;
  pair<size_t, T> p;
  for (size_t i = 0; i < a.NumRow(); ++i)
    u->SetElement(i, DotProduct(a.Row(i), v));
}

}  // namespace kernel

#endif  // KERNEL_LIB_SEMI_DENSE_MATRIX_H_
